@@ -553,7 +553,7 @@ charEventHandler(int wtime)
  * add primary menu
  */
     void
-gui_mch_add_menu_item(GuiMenu *menu, GuiMenu *parent, int idx)
+gui_mch_add_menu_item(VimMenu *menu, VimMenu *parent, int idx)
 {
     union myMenuItemUnion *menuItemUnion = NULL;
     struct IntuiText *menutext = NULL;
@@ -561,6 +561,10 @@ gui_mch_add_menu_item(GuiMenu *menu, GuiMenu *parent, int idx)
     assert(menu != NULL);
     assert(parent != NULL);
 
+    /* Don't add menu separator */
+    if (is_menu_separator(menu->name))
+	return;
+
     if (parent->menuItemPtr == NULL)
 	return;
 
@@ -649,12 +653,12 @@ getMenu(struct RastPort *rast, int left, STRPTR name)
  * add  1st level submenu item
  */
     void
-gui_mch_add_menu(GuiMenu *menu, GuiMenu *parent, int idx)
+gui_mch_add_menu(VimMenu *menu, VimMenu *parent, int idx)
 {
     struct Menu	*newMenu;
     int		pos = 0;
 
-    if (!gui_menubar_menu(menu->name))
+    if (!menubar_menu(menu->name))
 	return;
 
     menu->menuPtr = newMenu = getMenu(gui.window->RPort, 0, menu->dname);
@@ -780,6 +784,26 @@ gui_mch_exit(int returnCode)
     }
 }
 
+/*
+ * Get the position of the top left corner of the window.
+ */
+    int
+gui_mch_get_winpos(int *x, int *y)
+{
+    /* TODO */
+    return FAIL;
+}
+
+/*
+ * Set the position of the top left corner of the window to the given
+ * coordinates.
+ */
+    void
+gui_mch_set_winpos(int x, int y)
+{
+    /* TODO */
+}
+
     void
 gui_mch_set_winsize(int width, int height, int min_width, int min_height,
 		    int base_width, int base_height)
@@ -855,16 +879,19 @@ gui_mch_set_font(GuiFont font)
     /*D("gui_mch_set_font");*/
 }
 
+#if 0 /* not used */
     int
-gui_mch_same_font( GuiFont f1, GuiFont f2)
+gui_mch_same_font(GuiFont f1, GuiFont f2)
 {
     D("gui_mch_same_font");
 }
+#endif
 
     void
-gui_mch_free_font( GuiFont font)
+gui_mch_free_font(GuiFont font)
 {
-    D("gui_mch_free_font");
+    if (font)
+	D("gui_mch_free_font");
 }
 
 #define RGB(a,b,c) ((a&&0xff)*0x10000+(b*0xff)*0x100+(c&0xff))
@@ -1169,20 +1196,20 @@ gui_mch_set_menu_pos(int x, int y, int w, int h)
 }
 
     void
-gui_mch_destroy_menu(GuiMenu *menu)
+gui_mch_destroy_menu(VimMenu *menu)
 {
     D("gui_mch_destroy_menu");
     ClearMenuStrip(gui.window);
 }
 
     void
-gui_mch_menu_grey(GuiMenu *menu, int grey)
+gui_mch_menu_grey(VimMenu *menu, int grey)
 {
     D("gui_mch_menu_grey");
 }
 
     void
-gui_mch_menu_hidden(GuiMenu *menu, int hidden)
+gui_mch_menu_hidden(VimMenu *menu, int hidden)
 {
     D("gui_mch_menu_hidden");
     ClearMenuStrip(gui.window);
@@ -1252,7 +1279,7 @@ gui_mch_setmouse(x, y)
 }
 
     void
-gui_mch_show_popupmenu(GuiMenu *menu)
+gui_mch_show_popupmenu(VimMenu *menu)
 {
     /* TODO */
 }
@@ -1288,6 +1315,7 @@ gui_mch_get_lightness(GuiColor pixel)
     return (int)rc;
 }
 
+#if (defined(SYNTAX_HL) && defined(WANT_EVAL)) || defined(PROTO)
     char_u *
 gui_mch_get_rgb(GuiColor pixel)
 {
@@ -1300,6 +1328,6 @@ gui_mch_get_rgb(GuiColor pixel)
 
     return retval;
 }
-
+#endif
 
 #endif /* USE_AMIGA_GUI*/
