@@ -32,15 +32,21 @@
 #include <limits.h>
 #include <process.h>
 
-#define STRICT
-#define WIN32_LEAN_AND_MEAN
+#ifndef STRICT
+# define STRICT
+#endif
+#ifndef WIN32_LEAN_AND_MEAN
+# define WIN32_LEAN_AND_MEAN
+#endif
 #include <windows.h>
 
 #undef chdir
 #ifdef __GNUC__
-#include <dirent.h>
+# ifndef __MINGW32__
+#  include <dirent.h>
+# endif
 #else
-#include <direct.h>
+# include <direct.h>
 #endif
 
 
@@ -65,12 +71,14 @@ FILE* fdDump = NULL;
 # define DWORD int
 # define BOOL int
 # define LPSTR int
+# define LPTSTR int
 # define KEY_EVENT_RECORD int
 # define MOUSE_EVENT_RECORD int
 # define WINAPI
 # define CONSOLE_CURSOR_INFO int
 # define LPCSTR char_u *
 # define WINBASEAPI
+# define INPUT_RECORD int
 #endif
 
 #ifndef USE_GUI_WIN32
@@ -91,6 +99,7 @@ PFNGCKLN    s_pfnGetConsoleKeyboardLayoutName = NULL;
 #endif
 
 #if defined(__GNUC__) && !defined(PROTO)
+# ifndef __MINGW32__
 int _stricoll(char *a, char *b)
 {
     // the ANSI-ish correct way is to use strxfrm():
@@ -104,8 +113,9 @@ char * _fullpath(char *buf, char *fname, int len)
 {
     LPTSTR toss;
 
-   return (char *) GetFullPathName(fname, len, buf, &toss);
+    return (char *)GetFullPathName(fname, len, buf, &toss);
 }
+# endif
 
 int _chdrive(int drive)
 {
@@ -177,15 +187,17 @@ static void cursor_visible(BOOL fVisible);
 static BOOL write_chars(LPCSTR pchBuf, DWORD cchToWrite, DWORD* pcchWritten);
 static char_u tgetch(void);
 static void create_conin(void);
-static void mch_set_cursor_shape(int thickness);
 static int s_cursor_visible = TRUE;
 static int did_create_conin = FALSE;
+#else
+static int s_dont_use_vimrun = TRUE;
+static char *vimrun_path = "vimrun ";
 #endif
 
 /*
- * The new solution for opening a console doesn't work properly yet.
+ * The old solution for opening a console is still there.
  */
-#define OLD_CONSOLE_STUFF
+//#define OLD_CONSOLE_STUFF
 
 #ifdef OLD_CONSOLE_STUFF
 static void mch_open_console(void);
@@ -203,23 +215,28 @@ static int suppress_winsize = 1;	/* don't fiddle with console */
 # define VER_PLATFORM_WIN32_WINDOWS 1
 #endif
 
-static DWORD PlatformId(void);
+static void PlatformId(void);
 static DWORD g_PlatformId;
 
 /*
- *  Returns VER_PLATFORM_WIN32_NT (NT) or VER_PLATFORM_WIN32_WINDOWS (Win95)
+ * Set g_PlatformId to VER_PLATFORM_WIN32_NT (NT) or
+ * VER_PLATFORM_WIN32_WINDOWS (Win95).
  */
-    static DWORD
+    static void
 PlatformId(void)
 {
-    OSVERSIONINFO ovi;
+    static int done = FALSE;
 
-    ovi.dwOSVersionInfoSize = sizeof(ovi);
-    GetVersionEx(&ovi);
+    if (!done)
+    {
+	OSVERSIONINFO ovi;
 
-    g_PlatformId = ovi.dwPlatformId;
+	ovi.dwOSVersionInfoSize = sizeof(ovi);
+	GetVersionEx(&ovi);
 
-    return g_PlatformId;
+	g_PlatformId = ovi.dwPlatformId;
+	done = TRUE;
+    }
 }
 
 /*
@@ -324,6 +341,9 @@ const static struct
 
 /* The return code indicates key code size. */
     static int
+#ifdef __BORLANDC__
+    __stdcall
+#endif
 win32_kbd_patch_key(
     KEY_EVENT_RECORD* pker)
 {
@@ -334,7 +354,7 @@ win32_kbd_patch_key(
 
     if (s_iIsDead == 2)
     {
-#ifdef __GNUC__
+#if defined (__GNUC__) && !defined (__MINGW32__)
 	pker->AsciiChar = (CHAR) awAnsiCode[1];
 #else
 	pker->uChar.AsciiChar = (CHAR) awAnsiCode[1];
@@ -343,7 +363,7 @@ win32_kbd_patch_key(
 	return 1;
     }
 
-#ifdef __GNUC__
+#if defined (__GNUC__) && !defined (__MINGW32__)
     if (pker->AsciiChar != 0)
 #else
     if (pker->uChar.AsciiChar != 0)
@@ -379,7 +399,7 @@ win32_kbd_patch_key(
 			abKeystate, awAnsiCode, 0);
 
     if (s_iIsDead > 0)
-#ifdef __GNUC__
+#if defined (__GNUC__) && !defined (__MINGW32__)
 	pker->AsciiChar = (CHAR) awAnsiCode[0];
 #else
 	pker->uChar.AsciiChar = (CHAR) awAnsiCode[0];
@@ -436,11 +456,12 @@ decode_key_event(
 
     /* special cases */
     if ((nModifs & CTRL) != 0 && (nModifs & ~CTRL) == 0
-#ifdef __GNUC__
-	    && pker->AsciiChar == NUL)
+#if defined (__GNUC__) && !defined (__MINGW32__)
+	    && pker->AsciiChar == NUL
 #else
-	    && pker->uChar.AsciiChar == NUL)
+	    && pker->uChar.AsciiChar == NUL
 #endif
+	    )
     {
 	/* Ctrl-6 is Ctrl-^ */
 	if (pker->wVirtualKeyCode == '6')
@@ -495,7 +516,7 @@ decode_key_event(
     if (i < 0)
 	*pch = NUL;
     else
-#ifdef __GNUC__
+#if defined (__GNUC__) && !defined (__MINGW32__)
 	*pch = (i > 0)	?  pker->AsciiChar  :  NUL;
 #else
 	*pch = (i > 0)	?  pker->uChar.AsciiChar  :  NUL;
@@ -799,7 +820,7 @@ decode_mouse_event(
 #endif /* USE_MOUSE */
 
 
-#ifndef USE_GUI_WIN32	    /* this isn't used for the GUI */
+#ifdef MCH_CURSOR_SHAPE
 /*
  * Set the shape of the cursor.
  * 'thickness' can be from 1 (thin) to 99 (block)
@@ -831,6 +852,23 @@ mch_update_cursor(void)
 	thickness = cursor_table[idx].percentage;
     mch_set_cursor_shape(thickness);
 }
+#endif
+
+#ifndef USE_GUI_WIN32	    /* this isn't used for the GUI */
+/*
+ * Handle FOCUS_EVENT.
+ */
+    static void
+handle_focus_event(INPUT_RECORD ir)
+{
+    g_fJustGotFocus = ir.Event.FocusEvent.bSetFocus;
+    if (g_fJustGotFocus)
+	check_timestamps();
+#ifdef AUTOCMD
+    apply_autocmds(g_fJustGotFocus ? EVENT_FOCUSGAINED
+				: EVENT_FOCUSLOST, NULL, NULL, FALSE, curbuf);
+#endif
+}
 
 /*
  * Wait until console input from keyboard or mouse is available,
@@ -877,7 +915,7 @@ WaitForChar(long msec)
 	if (cRecords > 0)
 	{
 #ifdef MULTI_BYTE_IME
-	    /* Windows IME sends two '\n's with only one 'ENTER'. 
+	    /* Windows IME sends two '\n's with only one 'ENTER'.
 	       first, wVirtualKeyCode == 13. second, wVirtualKeyCode == 0 */
 	    if (ir.EventType == KEY_EVENT && ir.Event.KeyEvent.bKeyDown)
 	    {
@@ -896,10 +934,9 @@ WaitForChar(long msec)
 #endif
 
 	    ReadConsoleInput(g_hConIn, &ir, 1, &cRecords);
-#ifdef USE_CLIPBOARD
+
 	    if (ir.EventType == FOCUS_EVENT)
-		g_fJustGotFocus = ir.Event.FocusEvent.bSetFocus;
-#endif
+		handle_focus_event(ir);
 	    else if (ir.EventType == WINDOW_BUFFER_SIZE_EVENT)
 		set_winsize(Rows, Columns, FALSE);
 #ifdef USE_MOUSE
@@ -961,13 +998,8 @@ tgetch(void)
 	    if (decode_key_event(&ir.Event.KeyEvent, &ch, &g_chPending, TRUE))
 		return ch;
 	}
-#ifdef USE_CLIPBOARD
 	else if (ir.EventType == FOCUS_EVENT)
-	{
-	    g_fJustGotFocus = ir.Event.FocusEvent.bSetFocus;
-	    /* TRACE("tgetch: Focus %d\n", g_fJustGotFocus); */
-	}
-#endif
+	    handle_focus_event(ir);
 	else if (ir.EventType == WINDOW_BUFFER_SIZE_EVENT)
 	{
 	    set_winsize(Rows, Columns, FALSE);
@@ -999,20 +1031,47 @@ mch_inchar(
     long	time)
 {
 #ifndef USE_GUI_WIN32	    /* this isn't used for the GUI */
+
     int		len = 0;
     int		c;
 
+    static int	once_already = 0;
+
     if (time >= 0)
     {
 	if (!WaitForChar(time))     /* no character available */
+	{
+	    once_already = 0;
 	    return 0;
+	}
     }
     else    /* time == -1, wait forever */
     {
 	/* If there is no character available within 2 seconds (default),
 	 * write the autoscript file to disk */
-	if (!WaitForChar(p_ut))
+	if (once_already == 2)
 	    updatescript(0);
+	else if (once_already == 1)
+	{
+	    setcursor();
+	    once_already = 2;
+	    return 0;
+	}
+	else if (!WaitForChar(p_ut))
+	{
+#ifdef AUTOCMD
+	    if (has_cursorhold() && get_real_state() == NORMAL_BUSY)
+	    {
+		apply_autocmds(EVENT_CURSORHOLD, NULL, NULL, FALSE, curbuf);
+		update_screen(VALID);
+
+		once_already = 1;
+		return 0;
+	    }
+	    else
+#endif
+		updatescript(0);
+	}
     }
 
     /*
@@ -1076,6 +1135,7 @@ mch_inchar(
     }
 #endif
 
+    once_already = 0;
     return len;
 #else /* USE_GUI_WIN32 */
     return 0;
@@ -1083,21 +1143,90 @@ mch_inchar(
 }
 
 #ifdef USE_GUI_WIN32
+
+#ifndef __MINGW32__
+# include <shellapi.h>	/* required for FindExecutable() */
+#endif
+
+    static int
+executable_exists(char *name)
+{
+    char vimrun_location[2 * _MAX_PATH + 2];
+    char widename[2 * _MAX_PATH];
+
+    /* There appears to be a bug in FindExecutableA() on Windows NT.
+     * Use FindExecutableW() instead... */
+    if (g_PlatformId == VER_PLATFORM_WIN32_NT)
+    {
+	MultiByteToWideChar(CP_ACP, 0, (LPCTSTR)name, -1,
+						 (LPWSTR)widename, _MAX_PATH);
+	if (FindExecutableW((LPCWSTR)widename, (LPCWSTR)"",
+				     (LPWSTR)vimrun_location) > (HINSTANCE)32)
+	    return TRUE;
+    }
+    else
+    {
+	if (FindExecutableA((LPCTSTR)name, (LPCTSTR)"",
+				     (LPTSTR)vimrun_location) > (HINSTANCE)32)
+	    return TRUE;
+    }
+    return FALSE;
+}
+
 /*
  * GUI version of mch_windinit().
  */
     void
 mch_windinit()
 {
+#ifndef __MINGW32__
     extern int _fmode;
+#endif
 
     PlatformId();
 
+    /* Let critical errors result in a failure, not in a dialog box.  Required
+     * for the timestamp test to work on removed floppies. */
+    SetErrorMode(SEM_FAILCRITICALERRORS);
+
+    _fmode = O_BINARY;		/* we do our own CR-LF translation */
+
     /* Specify window size.  Is there a place to get the default from? */
     Rows = 25;
     Columns = 80;
 
-    _fmode = O_BINARY;		/* we do our own CR-LF translation */
+    /* Look for 'vimrun' */
+    if (!gui_is_win32s())
+    {
+	char_u vimrun_location[_MAX_PATH + 2];
+
+	/* First try in same directory as gvim.exe */
+	STRCPY(vimrun_location, exe_name);
+	STRCPY(gettail(vimrun_location), "vimrun.exe");
+	if (mch_getperm(vimrun_location) >= 0)
+	{
+	    STRCPY(gettail(vimrun_location), "vimrun ");
+	    vimrun_path = (char *)vim_strsave(vimrun_location);
+	    s_dont_use_vimrun = FALSE;
+	}
+	else if (executable_exists("vimrun.exe"))
+	    s_dont_use_vimrun = FALSE;
+
+	if (s_dont_use_vimrun)
+	    MessageBox(NULL,
+			"VIMRUN.EXE not found in your $PATH.\n"
+			"External commands will not pause after completion.\n"
+			"See  :help win32-vimrun  for more information.",
+			"Vim Warning",
+			MB_ICONWARNING);
+    }
+
+    /*
+     * If "finstr.exe" doesn't exist, use "grep -n" for 'grepprg'.
+     * Otherwise the default "findstr /n" is used.
+     */
+    if (!executable_exists("findstr.exe"))
+	set_option_value((char_u *)"grepprg", 0, (char_u *)"grep -n");
 
 #ifdef USE_CLIPBOARD
     clip_init(TRUE);
@@ -1154,10 +1283,16 @@ static int g_nOldColumns = 0;
 mch_windinit()
 {
     CONSOLE_SCREEN_BUFFER_INFO csbi;
+#ifndef __MINGW32__
     extern int _fmode;
+#endif
 
     PlatformId();
 
+    /* Let critical errors result in a failure, not in a dialog box.  Required
+     * for the timestamp test to work on removed floppies. */
+    SetErrorMode(SEM_FAILCRITICALERRORS);
+
     _fmode = O_BINARY;		/* we do our own CR-LF translation */
     out_flush();
 
@@ -1172,6 +1307,10 @@ mch_windinit()
     /* Get current text attributes */
     GetConsoleScreenBufferInfo(g_hSavOut, &csbi);
     g_attrCurrent = g_attrDefault = csbi.wAttributes;
+    if (cterm_normal_fg_color == 0)
+	cterm_normal_fg_color = (g_attrCurrent & 0xf) + 1;
+    if (cterm_normal_bg_color == 0)
+	cterm_normal_bg_color = ((g_attrCurrent >> 4) & 0xf) + 1;
     GetConsoleCursorInfo(g_hSavOut, &g_cci);
 
     /* set termcap codes to current text attributes */
@@ -1239,9 +1378,12 @@ mch_windexit(
     int r)
 {
     stoptermcap();
-    out_char('\r');
-    out_char('\n');
-    out_flush();
+    if (full_screen)
+    {
+	out_char('\r');
+	out_char('\n');
+	out_flush();
+    }
 
     if (g_fWindInitCalled)
 	settmode(TMODE_COOK);
@@ -1261,7 +1403,9 @@ mch_windexit(
 
     if (g_fWindInitCalled)
     {
+#ifdef WANT_TITLE
 	mch_restore_title(3);
+#endif
 
 #ifdef MCH_WRITE_DUMP
 	if (fdDump)
@@ -1409,6 +1553,7 @@ fname_case(
 }
 
 
+#ifdef WANT_TITLE
 /*
  * mch_settitle(): set titlebar of our window
  * Can the icon also be set?
@@ -1462,6 +1607,7 @@ mch_can_restore_icon()
 {
     return FALSE;
 }
+#endif /* WANT_TITLE */
 
 
 /*
@@ -1547,15 +1693,19 @@ mch_FullName(
     int len,
     int force)
 {
-    int nResult = FAIL;
+    int		nResult = FAIL;
 
     if (fname == NULL)		/* always fail */
 	return FAIL;
 
-    if (_fullpath(buf, fname, len) == NULL)
+    if (_fullpath(buf, fname, len - 1) == NULL)
 	STRNCPY(buf, fname, len);   /* failed, use the relative path name */
     else
+    {
+	if (mch_isdir(fname))
+	    STRCAT(buf, "\\");
 	nResult = OK;
+    }
 
     fname_case(buf);
 
@@ -1572,6 +1722,10 @@ mch_isFullName(
 {
     char szName[_MAX_PATH+1];
 
+    /* A name like "d:/foo" is always absolute */
+    if (fname[0] && fname[1] == ':' && (fname[2] == '/' || fname[2] == '\\'))
+	return TRUE;
+
     mch_FullName(fname, szName, _MAX_PATH, FALSE);
 
     return strcoll(fname, szName) == 0;
@@ -2097,7 +2251,7 @@ mch_display_error()
 static BOOL fUseConsole = TRUE;
 
     static int
-mch_system(char * cmd, int options)
+mch_system(char *cmd, int options)
 {
     STARTUPINFO		si;
     PROCESS_INFORMATION pi;
@@ -2107,7 +2261,16 @@ mch_system(char * cmd, int options)
     si.lpReserved = NULL;
     si.lpDesktop = NULL;
     si.lpTitle = NULL;
-    si.dwFlags = 0;
+    si.dwFlags = STARTF_USESHOWWINDOW;
+    /*
+     * It's nicer to run a filter command in a minimized window, but in
+     * Windows 95 this makes the command MUCH slower.  We can't do it under
+     * Win32s either as it stops the synchronous spawn workaround working.
+     */
+    if ((options & SHELL_DOOUT) && !mch_windows95() && !gui_is_win32s())
+	si.wShowWindow = SW_SHOWMINIMIZED;
+    else
+	si.wShowWindow = SW_SHOWNORMAL;
     si.cbReserved2 = 0;
     si.lpReserved2 = NULL;
 
@@ -2135,9 +2298,9 @@ mch_system(char * cmd, int options)
 #else
     /* There is a strange error on Windows 95 when using "c:\\command.com".
      * When the "c:\\" is left out it works OK...? */
-    if (STRNICMP(cmd, "c:", 2) == 0
-	    && (cmd[2] == '/' || cmd[2] == '\\')
-	    && STRNICMP(cmd + 3, "command.com", 11) == 0)
+    if (mch_windows95()
+	    && (STRNICMP(cmd, "c:/command.com", 14) == 0
+		|| STRNICMP(cmd, "c:\\command.com", 14) == 0))
 	cmd += 3;
 #endif
 
@@ -2226,10 +2389,11 @@ mch_call_shell(
 
 #ifndef USE_GUI_WIN32
     /*
-     * ALWAYS switch to non-termcap mode, otherwise ":r !ls" may crash.
+     * Switch to non-termcap mode, otherwise ":r !ls" may crash.
+     * The only exception is "!start".
      */
-    if (g_hCurOut == g_hConOut &&
-	    ((cmd == NULL) || STRNICMP(cmd, "start ", 6) != 0))
+    if (g_hCurOut == g_hConOut
+	    && (cmd == NULL || STRNICMP(cmd, "start ", 6) != 0))
     {
 	termcap_mode_end();
 	stopped_termcap_mode = TRUE;
@@ -2249,7 +2413,7 @@ mch_call_shell(
      * CTRL-C, Ctrl-Break or illegal instruction  might otherwise kill us.
      */
     signal(SIGINT, SIG_IGN);
-#ifdef __GNUC__
+#if defined(__GNUC__) && !defined(__MINGW32__)
     signal(SIGKILL, SIG_IGN);
 #else
     signal(SIGBREAK, SIG_IGN);
@@ -2265,18 +2429,21 @@ mch_call_shell(
 
     if (cmd == NULL)
     {
-	/* Set the SHELL_DOOUT flag, to avoid the "hit any key" prompt */
-	x = mch_system(p_sh, options | SHELL_DOOUT);
+	x = mch_system(p_sh, options);
     }
     else
     {
 	/* we use "command" or "cmd" to start the shell; slow but easy */
 	char_u *newcmd;
 
-	newcmd = lalloc(STRLEN(p_sh) + STRLEN(p_shcf) + STRLEN(cmd) + 3, TRUE);
+	newcmd = lalloc(
+#ifdef USE_GUI_WIN32
+		STRLEN(vimrun_path) +
+#endif
+		STRLEN(p_sh) + STRLEN(p_shcf) + STRLEN(cmd) + 10, TRUE);
 	if (newcmd != NULL)
 	{
-	    if (STRNICMP(cmd, "start ", 6) == 0)
+	    if (STRNICMP(*cmd == '"' ? cmd + 1 : cmd, "start ", 6) == 0)
 	    {
 		STARTUPINFO		si;
 		PROCESS_INFORMATION	pi;
@@ -2289,6 +2456,8 @@ mch_call_shell(
 		si.cbReserved2 = 0;
 		si.lpReserved2 = NULL;
 		sprintf((char *)newcmd, "%s\0", cmd+6);
+		if (*cmd == '"')
+		    *newcmd = '"';
 		/*
 		 * Now, start the command as a process, so that it doesn't
 		 * inherit our handles which causes unpleasant dangling swap
@@ -2306,22 +2475,30 @@ mch_call_shell(
 			&pi))			// Process information
 		    x = 0;
 		else
+		{
 		    x = -1;
+#ifdef USE_GUI_WIN32
+		    EMSG("Command not found");
+#endif
+		}
 		/* Close the handles to the subprocess, so that it goes away */
 		CloseHandle(pi.hThread);
 		CloseHandle(pi.hProcess);
 	    }
 	    else
 	    {
-		sprintf((char *)newcmd, "%s %s %s", p_sh,
-#ifdef OLD_CONSOLE_STUFF
-			p_shcf,
+		sprintf((char *)newcmd, "%s%s %s %s",
+#if defined(OLD_CONSOLE_STUFF) || !defined(USE_GUI_WIN32)
+			"",
 #else
-		    /* Use "/k" when the window should stick around after the
-		     * command finished, but only when 'shcf' is "/c" */
-			((options & SHELL_DOOUT) || p_shcf[0] != '/'
-					 || p_shcf[1] != 'c') ? p_shcf : "/k",
+			/*
+			 * Do we need to use "vimrun"?
+			 */
+			((options & SHELL_DOOUT) || s_dont_use_vimrun) ?
+			    "" : vimrun_path,
 #endif
+			p_sh,
+			p_shcf,
 			cmd);
 		x = mch_system((char *)newcmd, options);
 	    }
@@ -2340,10 +2517,12 @@ mch_call_shell(
 	smsg("%d returned", x);
 	msg_putchar('\n');
     }
+#ifdef WANT_TITLE
     resettitle();
+#endif
 
     signal(SIGINT, SIG_DFL);
-#ifdef __GNUC__
+#if defined(__GNUC__) && !defined(__MINGW32__)
     signal(SIGKILL, SIG_DFL);
 #else
     signal(SIGBREAK, SIG_DFL);
@@ -2370,7 +2549,11 @@ mch_call_shell(
 mch_has_wildcard(
     char_u *s)
 {
+#ifdef VIM_BACKTICK
+    return (vim_strpbrk(s, (char_u *)"?*$`~") != NULL);
+#else
     return (vim_strpbrk(s, (char_u *)"?*$~") != NULL);
+#endif
 }
 
 
@@ -2393,10 +2576,11 @@ pstrcmp(
  * Recursively build up a list of files in "gap" matching the first wildcard
  * in `path'.  Called by expand_wildcards().
  */
-    int
-mch_expandpath(
+    static int
+win32_expandpath(
     struct growarray	*gap,
     char_u		*path,
+    char_u		*wildc,
     int			flags)
 {
     char		buf[_MAX_PATH+1];
@@ -2406,6 +2590,7 @@ mch_expandpath(
     HANDLE		hFind;
     int			matches;
     int			start_dot_ok;
+    int			len;
 
     start_len = gap->ga_len;
 
@@ -2414,26 +2599,26 @@ mch_expandpath(
      * Copy it into `buf', including the preceding characters.
      */
     p = buf;
-    s = NULL;
+    s = buf;
     e = NULL;
     while (*path)
     {
-	if (*path == '\\' || *path == ':' || *path == '/')
+	if (path >= wildc && rem_backslash(path))	/* remove a backslash */
+	    ++path;
+	else if (*path == '\\' || *path == ':' || *path == '/')
 	{
 	    if (e)
 		break;
 	    else
-		s = p;
+		s = p + 1;
 	}
-	if (*path == '*' || *path == '?')
+	else if (*path == '*' || *path == '?')
 	    e = p;
 	*p++ = *path++;
     }
     e = p;
-    if (s)
-	s++;
-    else
-	s = buf;
+    *e = NUL;
+    /* now we have one wildcard component between `s' and `e' */
 
 #ifdef USE_GUI_WIN32
     if (gui_is_win32s())
@@ -2448,13 +2633,11 @@ mch_expandpath(
 	{
 	    *e++ = '.';
 	    *e++ = '*';
+	    *e = NUL;
 	}
     }
 #endif
 
-    /* now we have one wildcard component between `s' and `e' */
-    *e = NUL;
-
     start_dot_ok = (buf[0] == '.' || buf[0] == '*');
 
     /* If we are expanding wildcards, we try both files and directories */
@@ -2475,11 +2658,22 @@ mch_expandpath(
 			    && (s[1] != '.' || s[2] != NUL)))
 		    && (*path == NUL || mch_isdir(buf)))
 	    {
-		STRCAT(buf, path);
+		len = STRLEN(buf);
+		STRCPY(buf + len, path);
 		if (mch_has_wildcard(path))	    /* handle more wildcards */
-		    (void)mch_expandpath(gap, buf, flags);
+		{
+		    /* need to expand another component of the path */
+		    /* remove backslashes for the remaining components only */
+		    (void)win32_expandpath(gap, buf, buf + len + 1, flags);
+		}
 		else if (mch_getperm(buf) >= 0)	    /* add existing file */
-		    addfile(gap, buf, flags);
+		{
+		    /* no more wildcards, check if there is a match */
+		    /* remove backslashes for the remaining components only */
+		    backslash_halve(buf + len + 1, FALSE);
+		    if (mch_getperm(buf) >= 0)	/* add existing file */
+			addfile(gap, buf, flags);
+		}
 	    }
 
 	    c = FindNextFile(hFind, &fb);
@@ -2493,6 +2687,14 @@ mch_expandpath(
     return matches;
 }
 
+    int
+mch_expandpath(
+    struct growarray	*gap,
+    char_u		*path,
+    int			flags)
+{
+    return win32_expandpath(gap, path, path, flags);
+}
 
 /*
  * The normal _chdir() does not change the default drive.  This one does.
@@ -2611,6 +2813,7 @@ termcap_mode_start(void)
     g_coordOrig.Y = csbi.dwCursorPosition.Y;
 
     if (g_hConOut == INVALID_HANDLE_VALUE)
+    {
 	/* Create a new screen buffer in which we do all of our editing.
 	 * This means we can restore the original screen when we finish. */
 	g_hConOut = CreateConsoleScreenBuffer(GENERIC_WRITE | GENERIC_READ,
@@ -2618,11 +2821,14 @@ termcap_mode_start(void)
 					      CONSOLE_TEXTMODE_BUFFER,
 					      (LPVOID) NULL);
 
-    coord.X = coord.Y = 0;
-    FillConsoleOutputCharacter(g_hConOut, ' ', Rows * Columns, coord, &dwDummy);
-    FillConsoleOutputAttribute(g_hConOut, wAttr, Rows*Columns, coord, &dwDummy);
-
-    copy_screen_buffer_text(g_hSavOut, g_hConOut);
+	/* initialize the new screenbuffer with what is in the current one */
+	coord.X = coord.Y = 0;
+	FillConsoleOutputCharacter(g_hConOut, ' ',
+					     Rows * Columns, coord, &dwDummy);
+	FillConsoleOutputAttribute(g_hConOut, wAttr,
+					     Rows * Columns, coord, &dwDummy);
+	copy_screen_buffer_text(g_hSavOut, g_hConOut);
+    }
 
     g_hCurOut = g_hConOut;
     SetConsoleActiveScreenBuffer(g_hCurOut);
@@ -2631,7 +2837,9 @@ termcap_mode_start(void)
     set_scroll_region(0, 0, Columns - 1, Rows - 1);
     check_winsize();
 
+#ifdef WANT_TITLE
     resettitle();
+#endif
 
     textattr(wAttr);
 }
@@ -2641,11 +2849,17 @@ termcap_mode_start(void)
  * Turn off termcap mode by restoring the screen buffer we had upon startup
  */
     static void
-termcap_mode_end()
+termcap_mode_end(void)
 {
     g_attrSave = g_attrCurrent;
 
     ResizeConBufAndWindow(g_hCurOut, g_nOldColumns, g_nOldRows);
+
+    /* This weird Sleep(0) is required to allow Windows to really resize the
+     * console window.  Apparently it's done asynchronously, which may cause
+     * the following screenbuffer switch to go wrong */
+    Sleep(0);
+
     check_winsize();
 
     g_hCurOut = g_hSavOut;
@@ -2665,7 +2879,9 @@ termcap_mode_end()
     clear_chars(g_coordOrig,
 		g_srOrigWindowRect.Right - g_srOrigWindowRect.Left + 1);
 
+#ifdef WANT_TITLE
     mch_restore_title(3);
+#endif
 
     SetConsoleMode(g_hConIn,  g_cmodein);
     SetConsoleMode(g_hSavOut, g_cmodeout);
@@ -3013,7 +3229,9 @@ cursor_visible(
     BOOL fVisible)
 {
     s_cursor_visible = fVisible;
+#ifdef MCH_CURSOR_SHAPE
     mch_update_cursor();
+#endif
 }
 
 
@@ -3072,7 +3290,7 @@ mch_write(
 
     if (!term_console)
     {
-	write(1, s, (unsigned) len);
+	write(1, s, (unsigned)len);
 	return;
     }
 
@@ -3399,12 +3617,17 @@ mch_breakcheck()
 
 /*
  * How much memory is available?
+ * Return sum of available physical and page file memory.
  */
     long_u
 mch_avail_mem(
     int special)
 {
-    return LONG_MAX;	    /* virtual memory, eh? */
+    MEMORYSTATUS	ms;
+
+    ms.dwLength = sizeof(MEMORYSTATUS);
+    GlobalMemoryStatus(&ms);
+    return(ms.dwAvailPhys + ms.dwAvailPageFile);
 }
 
 
@@ -3434,6 +3657,61 @@ mch_screenmode(
     return FAIL;
 }
 
+#ifdef WANT_EVAL
+/*
+ * Call a DLL routine which takes either a string or int param
+ * and returns an allocated string.
+ */
+typedef LPTSTR (*MYSTRPROC)(LPTSTR);
+typedef LPTSTR (*MYINTPROC)(int);
+
+    char_u *
+mch_libcall(
+    char_u  *libname,
+    char_u  *funcname,
+    char_u  *argstring,
+    int	    argint)
+{
+    HINSTANCE hinstLib;
+    MYSTRPROC ProcAdd;
+    MYINTPROC ProcAddI;
+    char_u  *retval = NULL;
+
+    BOOL fFreeResult, fRunTimeLinkSuccess = FALSE;
+
+    // Get a handle to the DLL module.
+    hinstLib = LoadLibrary(libname);
+
+    // If the handle is valid, try to get the function address.
+    if (hinstLib != NULL)
+    {
+	if (argstring != NULL)
+	{
+	    ProcAdd = (MYSTRPROC) GetProcAddress(hinstLib, funcname);
+	    if (fRunTimeLinkSuccess = (ProcAdd != NULL))
+		retval = (ProcAdd) (argstring);
+	}
+	else
+	{
+	    ProcAddI = (MYINTPROC) GetProcAddress(hinstLib, funcname);
+	    if (fRunTimeLinkSuccess = (ProcAddI != NULL))
+		retval = (ProcAddI) (argint);
+	}
+
+	// save the string before we free the library
+	if (retval != NULL)
+	    retval = vim_strsave(retval);
+
+	// Free the DLL module.
+	fFreeResult = FreeLibrary(hinstLib);
+    }
+
+    if (!fRunTimeLinkSuccess)
+	EMSG("Library call failed");
+
+    return retval;
+}
+#endif
 
 /*
  * mch_rename() works around a bug in rename (aka MoveFile) in
@@ -3610,20 +3888,31 @@ clip_mch_request_selection()
 
 	if (hMem != NULL && str != NULL)
 	{
-	    /*successful lock - must unlock when finished*/
+	    /* successful lock - must unlock when finished */
 	    if (*str != NUL)
 	    {
-		LPCSTR psz = (LPCSTR)str;
-		char_u *temp_clipboard = (char_u *)lalloc(STRLEN(psz) + 1, TRUE);
-		char_u *pszTemp = temp_clipboard;
+		LPCSTR		psz = (LPCSTR)str;
+		char_u		*temp_clipboard;
+		char_u		*pszTemp;
+		const char	*pszNL;
+		int		len;
 
+		temp_clipboard = (char_u *)lalloc(STRLEN(psz) + 1, TRUE);
 		if (temp_clipboard != NULL)
 		{
+		    /* Translate <CR><NL> into <NL>. */
+		    pszTemp = temp_clipboard;
 		    while (*psz != NUL)
 		    {
-			const char* pszNL = strchr(psz, '\r');
-			const int len = (pszNL != NULL) ?
-					    pszNL - psz : STRLEN(psz);
+			pszNL = psz;
+			for (;;)
+			{
+			    pszNL = strchr(pszNL, '\r');
+			    if (pszNL == NULL || pszNL[1] == '\n')
+				break;
+			    ++pszNL;
+			}
+			len = (pszNL != NULL) ?  pszNL - psz : STRLEN(psz);
 			STRNCPY(pszTemp, psz, len);
 			pszTemp += len;
 			if (pszNL != NULL)
@@ -3632,11 +3921,11 @@ clip_mch_request_selection()
 		    }
 		    *pszTemp = NUL;
 		    clip_yank_selection(type, temp_clipboard,
-					(long)(pszTemp - temp_clipboard));
+					    (long)(pszTemp - temp_clipboard));
 		    vim_free(temp_clipboard);
 		}
 	    }
-	    /*unlock the global object*/
+	    /* unlock the global object */
 	    (void)GlobalUnlock(hMem);
 	}
 	CloseClipboard();
